# distortion_test.R
# RT 20.07.2015
# An R script to measure distortion generated by filters in the rt-plugins collection.
# This generates a test tone (modulated sine, dithered) and applies a given filter.
# The spectrum is plotted before and after; distortion (e.g. due to round-off in filter
# coefficients and/or re-quantization) shows up as spurious tones.  If all is well, these tones
# should be 

rm(list=ls());
graphics.off();

# LADSPA filter to test
filter <- "RTparaeq,24,20,50";  # a hard one: 24dB boost at 20Hz, with high Q=50
filter <- "RThighpass,20,5";    # a hard one: high-Q highpass at a low 20Hz

# frequency and level of test signal (modulated sine):
freq <- 80;          # [Hz]
#siglevel <- 2^-14;   # 1.0 = 0dBFS; 2^-14 = -84dBFS
siglevel <- 0.8;

# output should be dithered, otherwise distortion products could be due
# just to re-quantization at output:
ditheroutput <- TRUE;

stimfile <- "modsine.wav";
respfile <- "modsine_resp.wav";

# generate the modulated sine test signal =============================================
cat("\ngenerating modulated sine...\n");
source("./generate_mod_sine.R");
generate_mod_sine( freq, 6.0, 44100, bitdepth=16, level=siglevel, filename=stimfile,
                   dither=TRUE );

# apply our filter ====================================================================
if (ditheroutput) {  # output in 32-bit float then dither down to 16-bit:
  cmd <- sprintf( "ecasound -q -x -i %s -el:%s -f:f32_le,1,44100 -o stdout | sox -q -c 1 -r 44100 -b 32 -e float -L -t raw - -e signed -c 1 -b 16 -t wav %s dither", stimfile, filter, respfile);
} else {
  cmd <- sprintf("ecasound -x -q -i %s -el:%s -f:16,1,44100 -o %s", stimfile, filter, respfile);
}

cat("\nrunning filter:\n  ",cmd,"\n");
system(cmd);

# plot input spectrum =================================================================
cat("\ncalculating input spectrum...\n");
require(tuneR);
s <- readWave( stimfile );
y <- s@left / 2^15;
cat( "loaded", length(y), "samples from", stimfile, "\n" );
source("./spectrum_windowed.R");
x11(width=12,height=4);
my.spec <- spectrum.windowed( y, fftlen=2^16, xlim=c(5,20000),
           main="Test Tone Spectrum", max0=FALSE, log="x", lwd=2 );
abline(h=20*log10(2^(-16)),col="grey");
abline( h=-96, col="grey");  # show -96dBFS for reference

# plot output spectrum =================================================================
cat("\ncalculating output spectrum...\n");
s <- readWave( respfile );
y <- s@left / 2^15;
cat( "loaded", length(y), "samples from", respfile, "\n" );
x11(width=12,height=4);
my.spec <- spectrum.windowed( y, fftlen=2^16, max0=FALSE, xlim=c(5,20000), log="x", lwd=2,
           main=sprintf("Output Spectrum\n %s (%s)",filter, ifelse(ditheroutput,"TPDF dithered","no dither")) );
abline( h=-96, col="grey");  # show -96dBFS for reference

